import pandas as pd
import numpy as np
from rdkit import Chem
from rdkit.Chem import PandasTools

from __test_skeleton import BasicTest
import mol_conversion
from df_ops import nan_ops
from df_ops import column_ops
from df_ops import header_ops
from df_ops import duplicates
from df_ops import sdf_tools


class NanTests(BasicTest):
    def __init__(self) -> None:
        self.test_df = pd.DataFrame({'a': ['', '5', '6'], 'b': ['a', '8', 'None']})
        self.blank_solution = pd.DataFrame({'a': [np.nan, '5', '6'], 'b': ['a', '8', np.nan]})
        self.non_num_solution = pd.DataFrame({'a': [np.nan, '5', '6'], 'b': [np.nan, '8', np.nan]})
            
    def convert_to_nan_test(self):
        """Tests convert list of nas to nan."""
        
        # Test replace nan
        na_df = nan_ops.convert_to_nan(self.test_df)
        assert na_df.equals(self.blank_solution)
        
        # Test case sensitivity
        na_df = nan_ops.convert_to_nan(self.test_df, na=('A'))
        assert na_df.equals(pd.DataFrame({'a': ['', '5', '6'], 'b': [np.nan, '8', 'None']}))
        
    def non_num_to_nan(self):
        """Tests ability to convert all non-numeric values to np.nan."""
        
        # Test on only first column
        na_df = nan_ops.non_num_to_nan(self.test_df, self.test_df.columns[0])
        assert na_df.equals(pd.DataFrame({'a': [np.nan, '5', '6'], 'b': ['a', '8', 'None']}))
        
        # Test on only second column (unchanged)
        na_df = nan_ops.non_num_to_nan(self.test_df, self.test_df.columns[1])
        assert na_df.equals(self.test_df)
        
        # Test on both columns
        na_df = nan_ops.non_num_to_nan(self.test_df, self.test_df.columns)
        assert na_df.equals(self.non_num_solution)

    def nan_col_indices_test (self):
        """Tests nan index finder."""
        
        # Test each col separately
        assert nan_ops.nan_col_indices(self.blank_solution, self.blank_solution.columns[0]) == [0]
        assert nan_ops.nan_col_indices(self.blank_solution, self.blank_solution.columns[1]) == [2]
        
    def remove_nan_cols_test(self):
        nan_col_df = self.test_df.copy()
        
        # Append new columns
        nan_col_df['c'] = 3*['None']
        nan_col_df['d'] = 3*['nan']
        nan_col_df['e'] = 3*['']
        nan_col_df['f'] = 3*[np.nan]
        
        # Check new columns were added successfully
        assert len(nan_col_df.columns) == 6
        
        # Test blank column removal
        assert nan_ops.remove_nan_cols(nan_col_df).equals(self.test_df)
        
    def remove_nan_rows_test(self):
        self.non_num_solution = pd.DataFrame({'a': [np.nan, '5', '6'], 'b': [np.nan, '8', np.nan]})
        
        col_a = nan_ops.remove_nan_rows(self.blank_solution, ['a'])
        assert col_a.equals(pd.DataFrame({'a': ['5', '6'], 'b': ['8', np.nan]}))
        
        col_b = nan_ops.remove_nan_rows(self.blank_solution, ['b'])
        assert col_b.equals(pd.DataFrame({'a': [np.nan, '5'], 'b': ['a', '8']}))
        
        both_cols = nan_ops.remove_nan_rows(self.blank_solution, self.blank_solution.columns)
        assert both_cols.equals(pd.DataFrame({'a': ['5'], 'b': ['8']}))


class ColumnOpsTests(BasicTest):
    def __init__(self) -> None:
        self.test_df = pd.DataFrame({'a': [4, '5', '6'], 'b': ['6', '5', '4']})
        self.mol_test_df = pd.DataFrame({
            'SMILES': ['O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1', 
                       'O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1', 
                       'CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O',
                       None],
            'Molecule': [Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1'),
                         Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1'),
                         Chem.MolFromSmiles('CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O'),
                         None],
            'InChi': [Chem.MolToInchiKey(Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1')),
                      Chem.MolToInchiKey(Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1')),
                      Chem.MolToInchiKey(Chem.MolFromSmiles('CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O')),
                      None]
        })

    def drop_val_test(self):
        # Test number
        dropped1 = column_ops.drop_val(self.test_df, 'a', 4)
        assert dropped1.equals(pd.DataFrame({'a': ['5', '6'], 'b': ['5', '4']}))
        
        # Test string
        dropped2 = column_ops.drop_val(self.test_df, 'a', '4')
        assert dropped2.equals(self.test_df)
        
        # Test b column
        dropped3 = column_ops.drop_val(self.test_df, 'b', '4')
        assert dropped3.equals(pd.DataFrame({'a': [4, '5'], 'b': ['6', '5']}))
        
    def pull_val_test(self):
        # Test number
        pulled1 = column_ops.pull_val(self.test_df, 'a', 4)
        assert pulled1.equals(pd.DataFrame({'a': [4], 'b': ['6']}).astype(object))
        
        # Test string
        pulled2 = column_ops.pull_val(self.test_df, 'a', '6')
        assert pulled2.equals(pd.DataFrame({'a': ['6'], 'b': ['4']}, index=[2]))
        
        # Test b column
        pulled3 = column_ops.pull_val(self.test_df, 'b', '5')
        assert pulled3.equals(pd.DataFrame({'a': ['5'], 'b': ['5']}, index=[1]))
        
    def pull_not_val_test(self):
        # Test number
        pulled1 = column_ops.pull_not_val(self.test_df, 'a', 4)
        assert pulled1.equals(pd.DataFrame({'a': ['5', '6'], 'b': ['5', '4']}, index=[1, 2]).astype(object))
        
        # Test string
        pulled2 = column_ops.pull_not_val(self.test_df, 'a', '6')
        assert pulled2.equals(pd.DataFrame({'a': [4, '5'], 'b': ['6', '5']}))
        
        # Test b column
        pulled3 = column_ops.pull_not_val(self.test_df, 'b', '5')
        assert pulled3.equals(pd.DataFrame({'a': [4, '6'], 'b': ['6', '4']}, index=[0, 2]))
        
        # Test pull invalid number --> returns what was input
        pulled4 = column_ops.pull_not_val(self.test_df, 'b', '3')
        assert pulled4.equals(self.test_df)
        
    def smiles_to_mols_test(self):
        # Test with dropna
        out1 = column_ops.smiles_to_mols(self.mol_test_df, 'SMILES', 'ROMol')
        assert out1.ROMol.map(Chem.MolToSmiles, na_action='ignore').equals(
            self.mol_test_df.Molecule.map(Chem.MolToSmiles, na_action='ignore').dropna())
        
        # Test without dropna
        out2 = column_ops.smiles_to_mols(self.mol_test_df, 'SMILES', 'ROMol', dropna=False)
        assert out2.ROMol.map(Chem.MolToSmiles, na_action='ignore').equals(
            self.mol_test_df.Molecule.map(Chem.MolToSmiles, na_action='ignore'))
        
    def mols_to_inchis_test(self):
        # Test with dropna
        out1 = column_ops.mols_to_inchis(self.mol_test_df, 'Molecule', 'inchi')
        assert out1.inchi.equals(self.mol_test_df.InChi.dropna())
        
        # Test without dropna
        out2 = column_ops.mols_to_inchis(self.mol_test_df, 'Molecule', 'inchi', dropna=False)
        assert out2.inchi.equals(self.mol_test_df.InChi)
        
    def mols_to_smiles_test(self):
        # Test with dropna
        out1 = column_ops.mols_to_smiles(self.mol_test_df, 'Molecule', 'smiles')
        assert out1.smiles.equals(self.mol_test_df.SMILES.dropna())
        
        # Test without dropna
        out2 = column_ops.mols_to_smiles(self.mol_test_df, 'Molecule', 'smiles', dropna=False)
        assert out2.smiles.equals(self.mol_test_df.SMILES)
        
        
class HeaderOpsTests(BasicTest):
    def __init__(self) -> None:
        self.mol_test_df = pd.DataFrame({
            'SMILES': ['O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1', 
                       'O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1', 
                       'CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O',
                       'c'],
            'Molecule': [Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1'),
                         Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1'),
                         Chem.MolFromSmiles('CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O'),
                         None],
            'InChi': [Chem.MolToInchi(Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(Br)cc23)CCC1')),
                      Chem.MolToInchi(Chem.MolFromSmiles('O=C(O)C1(Sc2ccnc3ccc(C4CC4)cc23)CCC1')),
                      Chem.MolToInchi(Chem.MolFromSmiles('CC(C)(Sc1ccncc1-c1ccc(C#N)c2ccccc12)C(=O)O')),
                      None]
        })
        
    def id_nearest_col_test(self):
        # Test similar variations
        assert header_ops.id_nearest_col(self.mol_test_df, 'smiles') == 'SMILES'
        assert header_ops.id_nearest_col(self.mol_test_df, 'smIles') == 'SMILES'
        assert header_ops.id_nearest_col(self.mol_test_df, 'smes') == 'SMILES'
        
        # Test null return
        assert header_ops.id_nearest_col(self.mol_test_df, 's') == None
        
    def remove_header_chars_test(self):
        test_df1 = self.mol_test_df.copy()
        test_df2 = self.mol_test_df.copy()
        test_df3 = self.mol_test_df.copy()
        test_df1.columns = ['MILE', 'Molecule', 'InChi']
        test_df2.columns = ['SMLES', 'Molecule', 'nCh']
        test_df3.columns = ['SMILES', 'Molecule', 'InCh']
        
        # Test w/ case sensitive on
        assert header_ops.remove_header_chars(self.mol_test_df, 's').equals(test_df1)
        assert header_ops.remove_header_chars(self.mol_test_df, 'S').equals(test_df1)
        
        # Test w/ case sensitive off
        assert header_ops.remove_header_chars(self.mol_test_df, 's', case_insensitive=False).equals(self.mol_test_df)
        assert header_ops.remove_header_chars(self.mol_test_df, 'S', case_insensitive=False).equals(test_df1)
        
        # Test multiple instances
        assert header_ops.remove_header_chars(self.mol_test_df, 'i').equals(test_df2)
        assert header_ops.remove_header_chars(self.mol_test_df, 'i', case_insensitive=False).equals(test_df3)
        

class DuplicatesTests(BasicTest):
    def __init__(self) -> None:
        self.test_df = pd.DataFrame({'aa': ['5', '5', '6', 5, 6, '6'], 'b': [1, 2, 3, 4, 5, 6], 
                                     'c': ['a', 'a', 'b', 'b', 'b', 'b']})
        self.filt_not_dup_sol = pd.DataFrame({'aa': ['5', '6', 5, 6], 'b': [1, 3, 4, 5], 'c': ['a', 'b', 'b', 'b']}, 
                                             index=[0, 2, 3, 4])
        self.filt_dup_sol = pd.DataFrame({'aa': ['5', '6'], 'b': [2, 6], 'c': ['a', 'b']}, index=[1, 5])
        self.filt_multi_not_dup_sol = pd.DataFrame({'aa': ['5', '6'], 'b': [1, 3], 'c': ['a', 'b']}, 
                                                   index=[0, 2])
        self.filt_multi_dup_sol = pd.DataFrame({'aa': ['5', 5, 6, '6'], 'b': [2, 4, 5, 6], 
                                                'c': ['a', 'b', 'b', 'b']}, index=[1, 3, 4, 5])
        self.average_sol = pd.DataFrame({'aa': ['5', '6', 5, 6], 'b': [1.5, 4.5, 4, 5], 'c': ['a', 'b', 'b', 'b']}, 
                                        index=[0, 2, 3, 4])
    
    def filter_test(self):
        # Test filtering out duplicates
        assert duplicates.filter(self.test_df, ['aa']).equals(self.filt_not_dup_sol)
        
        # Test filtering out everything EXCEPT duplicates
        assert duplicates.filter(self.test_df, ['aa'], return_duplicates=True).equals(self.filt_dup_sol)
        
        # Test filtering across multiple subsets
        assert duplicates.filter(self.test_df, ['aa', 'c'], how='any').equals(self.filt_multi_not_dup_sol)
        assert duplicates.filter(self.test_df, ['aa', 'c'], how='any', return_duplicates=True).equals(
            self.filt_multi_dup_sol)

    def indices_test(self):
        assert duplicates.indices(self.test_df, 'c') == [1, 3, 4, 5]
        
    def remove_test(self):
        assert duplicates.remove(self.test_df, ['aa']).equals(self.filt_not_dup_sol)
        
    def average_test(self):
        assert duplicates.average(self.test_df, ['aa'], average_by='b').equals(self.average_sol) 
        
        
class SDFToolsTests(BasicTest):
    def __init__(self) -> None:
        self.excel_df = pd.read_excel('excel_test_case.xlsx')
        self.excel_df = column_ops.smiles_to_mols(self.excel_df, 'Smiles', 'Molecule')
        self.sdf_df = PandasTools.LoadSDF('sdf_test_case.sdf', molColName='Molecule')
        self.sdf_df.ID = list(range(4))
    
    def write_sdf_test(self):
        # Write new and load new
        sdf_tools.write_sdf(self.excel_df, 'sdf_test_out.sdf', mol_col_name='Molecule')
        load_sdf = PandasTools.LoadSDF('./sdf_test_out.sdf', molColName='Molecule')
        
        # Ensure Mols are the same
        assert mol_conversion.mols_2_smiles(load_sdf.Molecule) == \
            mol_conversion.mols_2_smiles(self.sdf_df.Molecule)
        
        # Ensure original SMILES are the same
        assert load_sdf.Smiles.equals(self.sdf_df.Smiles)
        
        
nan_tests = NanTests()
column_ops_tests = ColumnOpsTests()
header_ops_tests = HeaderOpsTests()
duplicates_tests = DuplicatesTests()
sdf_tools_tests = SDFToolsTests()

nan_tests.run_all_tests()
column_ops_tests.run_all_tests()
header_ops_tests.run_all_tests()
duplicates_tests.run_all_tests()
sdf_tools_tests.run_all_tests()



from naclo.mol_stats import mol_weights, carbon_num
from naclo.mol_conversion import *
from naclo.fragments import FragRemover
from naclo.neutralization import Neutralizer
from naclo.dataframes import Dataframes



import mol_conversion
from rdkit.Chem import PandasTools
import pandas as pd


def mols_2_smiles_test():
    assert sdf_smiles == mol_conversion.mols_2_smiles(sdf_mols)
    
def smiles_2_mols_test():
    assert excel_smiles == mol_conversion.mols_2_smiles(mol_conversion.smiles_2_mols(excel_smiles))
    
def mols_2_inchi_keys_test():
    assert excel_inchi_keys == mol_conversion.mols_2_inchi_keys(mol_conversion.smiles_2_mols(excel_smiles))
    
def smiles_2_inchi_keys_test():
    assert excel_inchi_keys == mol_conversion.smiles_2_inchi_keys(excel_smiles)


# Load excel test data
test_excel = pd.read_excel('excel_test_case.xlsx')
excel_smiles = list(test_excel.Smiles)
excel_inchi_keys = list(test_excel.InChi)

# Load SDF test data
test_sdf = PandasTools.LoadSDF('sdf_test_case.sdf', molColName='Molecule')
sdf_mols = list(test_sdf.Molecule)
sdf_smiles = list(test_sdf.Smiles)

# Run test functions
mols_2_smiles_test()
smiles_2_mols_test()
mols_2_inchi_keys_test()
smiles_2_inchi_keys_test()


from rdkit import Chem


# class Conversion:
def mols_2_smiles(mols):  # *
    """Generates SMILES strings from list of rdkit Mol objects.

    Args:
        mols (iter[rdkit Mol]): Contains RDKit Mols.

    Returns:
        list[str]: Contains SMILES strings.
    """
    return [Chem.MolToSmiles(mol) for mol in mols]

def smiles_2_mols(smiles):  # *
    """Generates rdkit Mol objects from SMILES strings.

    Args:
        smiles (iter[str]): Contains SMILES strings.

    Returns:
        list[rdkit Mol]: Contains RDKit Mols.
    """
    return [Chem.MolFromSmiles(smile) for smile in smiles]

def mols_2_inchi_keys(mols):  # *
    """Generates InChI key strings from rdkit Mol objects.

    Args:
        mols (iter[rdkit Mols]): Contains rdkit Mols.

    Returns:
        list[str]: Contains InChI key strings.
    """
    return [Chem.MolToInchiKey(mol) for mol in mols]

def smiles_2_inchi_keys(smiles):  # *
    """Generates InChI key strings from SMILES strings.

    Args:
        smiles (iter[str]): Contains SMILES strings.

    Returns:
        list[str]: Contains InChI key strings.
    """
    mols = smiles_2_mols(smiles)
    return mols_2_inchi_keys(mols)



from neutralization import *
import mol_conversion


def init_neutralization_rxns_test():
    rxns = init_neutralization_rxns()
    rxns = list(rxns.values()) +  list(rxns.keys())
    assert mol_conversion.mols_2_smiles(rxns) == ['O', '*']  # ????
    

def neutralize_charges_test():
    # Doesn't do anything yet
    pass


init_neutralization_rxns_test()


from rdkit import Chem
from rdkit.Chem.Descriptors import ExactMolWt
from naclo import mol_stats
from rdkit.Chem.SaltRemover import SaltRemover


class FragRemover():
    def mw(self, smile):  # *
        """Removes smaller fragments from a SMILES string by molecular weight.

        Args:
            smile (str): Mol SMILES representation.

        Returns:
            str: SMILES with fragments removed.
        """
        fragments = smile.split('.')

        # Calculate molecular weights of each fragment
        mws = [ExactMolWt(Chem.MolFromSmiles(frag)) for frag in fragments]

        # Locate the max molecular weight
        max_mw = max(mws)
        max_index = mws.index(max_mw)

        return fragments[max_index]
    
    def atom_count(self, smile):  # *
        """Removes smaller fragments from a SMILES string by atom count.

        Args:
            smile (str): Mol SMILES representation.

        Returns:
            str: SMILES with fragments removed.
        """
        fragments = smile.split('.')

        # Calculate atom counts of each fragment
        atom_counts = [Chem.MolFromSmiles(frag).GetNumAtoms() for frag in fragments]

        # Locate the max atom count
        max_atom_count = max(atom_counts)
        max_index = atom_counts.index(max_atom_count)

        return fragments[max_index]
    
    def carbon_count(self, smile):  # *
        """Removes smaller fragments from a SMILES string by carbon count.

        Args:
            smile (str): Mol SMILES representation.

        Returns:
            str: SMILES with fragments removed.
        """
        fragments = smile.split('.')
        
        # Calculate carbon counts of each fragment
        carbon_counts = [mol_stats.carbon_num(frag) for frag in fragments]
        
        # Locate the max carbon count
        max_carbon_count = max(carbon_counts)
        max_index = carbon_counts.index(max_carbon_count)
        
        return fragments[max_index]

    def remove_salts(self, mols, salts='[Cl,Br]'):  # *
        """Removes salts from iterable.

        Args:
            mols (iterable): Contains RDKit Mol objects.

        Returns:
            list: Contains RDKit Mol objects with salts removed.
        """
        remover = SaltRemover(defnData=salts)
            
        return [remover.StripMol(mol) for mol in mols]



from rdkit import Chem
from rdkit.Chem import AllChem


class Neutralizer:
    def init_neutralization_rxns(self, reactants_products={'[$([O-]);!$([O-][#7])]': 'O'}):
        """Builds reactions from dict of reactants and products.

        Args:
            reactants (dict, optional): Reactants (keys) and products (values) SMARTS subgroups. \
                Defaults to Carboxylic acids and alcohols ('[$([O-]);!$([O-][#7])]', 'O').

        Returns:
            dict: Contains Mol representations of reactants.
        """
        reactions = {}
        for reactant in reactants_products.keys():
            product = reactants_products[reactant]
            reactions[Chem.MolFromSmarts(reactant)] = Chem.MolFromSmarts(product)
            
        return reactions

    def neutralize_charges(self, mols, reactants_products={'[$([O-]);!$([O-][#7])]': 'O'}):
        """Neutralizes Mol charges by replacing reactants with products.

        Args:
            mols (list): RDKit Mol objects. Mols to neutralize.
            reactants_products (dict, optional): Reactants (keys) and products (values) SMARTS subgroups. \
                Defaults to Carboxylic acids and alcohols ('[$([O-]);!$([O-][#7])]', 'O').

        Returns:
            list: Neutralized RDKit Mols.
        """
        reactions = self.init_neutralization_rxns(reactants_products=reactants_products)
        
        neutralized_mols = []
        for mol in mols:
            # Iterate over neutralization reactions
            for (reactant, product) in zip(reactions.keys(), reactions.values()):
                # Loop until all instances have been found
                while mol.HasSubstructMatch(reactant):
                    mol = AllChem.ReplaceSubstructs(mol, reactant, product)[0]
                    mol.UpdatePropertyCache()     
            
            neutralized_mols.append(mol)
            
        return neutralized_mols



from __test_skeleton import BasicTest
from mol_stats import *
import mol_conversion


class MolStatsTests(BasicTest):
    def __init__(self) -> None:
        self.test_smile = 'CN=C=O'
        self.test_mols = mol_conversion.smiles_2_mols([self.test_smile, 'C'])
    
    def carbon_num_test(self):
        assert carbon_num(self.test_smile) == 2
    
    def mol_weights_test(self):
        assert list(map(lambda x: round(x, 3), mol_weights(self.test_mols))) == [57.021, 16.031]


mol_stats_tests = MolStatsTests()
mol_stats_tests.run_all_tests()



import fragments
from mol_conversion import smiles_2_mols, mols_2_smiles


frag_rm = fragments.FragRemover()

def frag_remover_test():
    # Test case
    test_smile = 'CCC.C.C'
    
    # Test all frag removing methods
    assert frag_rm.mw(test_smile) == 'CCC'
    assert frag_rm.atom_count(test_smile) == 'CCC'
    assert frag_rm.carbon_count(test_smile) == 'CCC'
    
def remove_salts_test():
    # Test case
    test_mols = smiles_2_mols(['CN(C)C.Cl', 'CN(C)C.N'])
    
    # Test default salts [Cl,Br]
    assert mols_2_smiles(fragments.remove_salts(test_mols)) == ['CN(C)C', 'CN(C)C.N']
    
    # Test custom salts [N]
    assert mols_2_smiles(fragments.remove_salts(test_mols, salts='[N]')) == ['CN(C)C.Cl', 'CN(C)C']


frag_remover_test()
remove_salts_test()



from rdkit import Chem
from warnings import warn
from rdkit.Chem import PandasTools


class Dataframes:
    def smiles_to_mols(self, df, smiles_name, mol_name, dropna=True):  # *
        """Adds rdkit Mol column to df using SMILES column as reference.

        Args:
            df (pandas DataFrame): DataFrame to add Mol column to.
            smiles_name (str): Name of SMILES column in df.
            molecule_name (str): Name of Mol column in df.
            dropna (bool, optional): Drop NA Mols. Defaults to True.

        Returns:
            pandas DataFrame: DataFrame with Mol column appended.
        """
        df[mol_name] = df[smiles_name].map(Chem.MolFromSmiles, na_action='ignore')  # Results in NA el if fails
        return df.dropna(subset=[mol_name]) if dropna else df

    def mols_to_inchis(self, df, mol_name, inchi_name, dropna=True):  # *
        """Adds InChi Key column to df using Mol column as reference.
        
        Args:
            df (pandas DataFrame): DataFrame to add InChi column to.
            mol_name (str): Name of InChi column in df.
            inchi_name (str): Name of InChi column in df.
            dropna (bool, optional): Drop NA InChis. Defaults to True.

        Returns:
            pandas DataFrame: DataFrame with InChi column appended.
        """
        df[inchi_name] = df[mol_name].map(Chem.MolToInchiKey, na_action='ignore')  # Results in NA el if fails
        return df.dropna(subset=[inchi_name]) if dropna else df
        
    def mols_to_smiles(self, df, mol_name, smiles_name, dropna=True):  # *
        """Adds SMILES Key column to df using Mol column as reference.

        Args:
            df (pandas DataFrame): DataFrame to add SMILES column to.
            mol_name (str): Name of Mol column in df.
            inchi_name (str): Name of InChi column in df.
            dropna (bool, optional): Drop NA SMILES. Defaults to True.

        Returns:
            pandas DataFrame: DataFrame with SMILES column appended.
        """
        df[smiles_name] = df[mol_name].map(Chem.MolToSmiles, na_action='ignore')  # Results in NA el if fails
        return df.dropna(subset=[smiles_name]) if dropna else df

    def write_sdf(self, df, out_path, mol_col_name, id_column_name='RowID'):  # *
        """Writes dataframe to SDF file. Includes ID name if ID is valid.

        Args:
            df (DataFrame): Data to write.
            out_path (str or file-like): Path to save SDF to.
            mol_col_name (str): Name of molecule column in dataframe.
            id_column_name (str, optional): Name of id column. Defaults to 'ID'.
        """
        try:
            PandasTools.WriteSDF(df, out_path, molColName=mol_col_name, properties=df.columns, idName=id_column_name)
        except KeyError:
            PandasTools.WriteSDF(df, out_path, molColName=mol_col_name, properties=df.columns, idName='RowID')
            warn(f'write_sdf \'{id_column_name}\' ID name invalid', UserWarning)
